#include <cstdlib>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include "UniformIntDistribution.hpp"
#include "hsvcolor.hpp"
#include "point.hpp"
#include "ray.hpp"
#include "rgbcolor.hpp"
#include "sphere.hpp"

#define SPHERES 20

HOST_DEV RGBColor get_color(Sphere* spheres, const Ray& ray) {
   Shade shade; double tmin; bool hit = false;
   for (unsigned int i = 0; i < SPHERES; ++i) {
      Shade last_shade; double last_tmin;
      if (spheres[i].hit(ray, last_tmin, last_shade)) {
	 if (!hit || last_tmin < tmin) {
	    tmin = last_tmin; shade = last_shade; hit = true;
	 }
      }
   }
   if (!hit) return RGBColor(0, 0, 0);
   double factor = normalize(ray.direction) * normalize(shade.normal);
   if (factor >= 0) return RGBColor(0, 0, 0); /* unexpected */
   return shade.color * fabs(factor);
}

template<typename T>
RGBColor get_rand_color(UniformIntDistribution<T>& prg) {
   float hue = prg.draw(360);
   return HSVColor(prg.draw(360), 0.9, 1).get_rgb();
}

int main(int argc, char** argv) {
   const char* cmdname = *argv++; --argc;
   if (argc != 1) {
      std::cerr << "Usage: " << cmdname << " out.jpg" << std::endl;
      exit(1);
   }
   const char* outfile = *argv++; --argc;

   UniformIntDistribution<int> prg;

   Sphere spheres[SPHERES];
   for (unsigned int i = 0; i < SPHERES; ++i) {
      for(;;) { /* until a non-intersecting sphere has been created */
	 Sphere candidate(
	    /* center = */ Point(-500 + prg.draw(1000),
	       -500 + prg.draw(1000), 500 + prg.draw(2000)),
	    /* radius = */ 50 + prg.draw(100),
	    /* color = */ get_rand_color(prg)
	 );
	 bool ok = true;
	 /* check if the candidate intersects with any previous sphere */
	 for (unsigned int j = 0; j < i; ++j) {
	    if (intersect(spheres[j], candidate)) {
	       ok = false; break;
	    }
	 }
	 if (ok) {
	    spheres[i] = candidate; break;
	 }
      }
   }

   unsigned int width = 2000; unsigned int height = 1000;
   double horizontal_diameter = 1000;
   double resolution = width / horizontal_diameter;
   double xmin = - horizontal_diameter / 2;
   double vertical_diameter = height / resolution;
   double ymin = - vertical_diameter / 2;

   GdkPixbuf* pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8,
      width, height);
   int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
   int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
   guchar* pixels = gdk_pixbuf_get_pixels(pixbuf);
   if (!pixels) {
      std::cerr << cmdname << ": failed to allocate pixbuf with " <<
	 width << " x " << height << " pixels" << std::endl;
      exit(1);
   }
   for (int i = 0; i < height; ++i) {
      for (int j = 0; j < width; ++j) {
	 guchar* pp = pixels + i * rowstride + j * n_channels;

	 double x = xmin + j/(double)resolution;
	 double y = ymin + i/(double)resolution;
	 Ray ray(Point(x, y, 0), Vector(0, 0, 1));
	 RGBColor color(get_color(spheres, ray));

	 pp[0] = color.red * 255;
	 pp[1] = color.green * 255;
	 pp[2] = color.blue * 255;
      }
   }

   gdk_pixbuf_save(pixbuf, outfile, "png", (GError **)0, (void*)0);
}
